package com.dreamcodex.ti;

import com.dreamcodex.ti.component.*;
import com.dreamcodex.ti.iface.IconProvider;
import com.dreamcodex.ti.iface.ScreenColorListener;
import com.dreamcodex.ti.iface.UndoRedoListener;
import com.dreamcodex.ti.util.ECMPalette;
import com.dreamcodex.ti.util.Globals;
import com.dreamcodex.ti.util.MutableFilter;
import com.dreamcodex.ti.util.TIGlobals;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.awt.image.FilteredImageSource;
import java.awt.image.ImageProducer;
import java.io.*;
import java.net.URL;
import java.util.*;

/**
 * Magellan
 * TI-99/4A graphical map editor
 *
 * @author Howard Kistler
 */

public class Magellan extends JFrame implements Runnable, WindowListener, ActionListener, MouseListener, MouseMotionListener, IconProvider, ScreenColorListener, UndoRedoListener {

// Constants -------------------------------------------------------------------------------/

    public static final boolean ISOMETRIC = false;

    public static final String VERSION_NUMBER = "2.0";

    public static final int COLOR_MODE_GRAPHICS_1 = 0;
    public static final int COLOR_MODE_BITMAP = 1;
    public static final int COLOR_MODE_ECM_2 = 2;
    public static final int COLOR_MODE_ECM_3 = 3;
    public static String[] COLOR_MODES = new String[] {
        "Graphics 1 Color Mode",
        "Bitmap Color Mode",
        "Enhanced Color Mode - 2 bpp",
        "Enhanced Color Mode - 3 bpp"
    };

    public static final int SCROLL_ORIENTATION_VERTICAL = 0;
    public static final int SCROLL_ORIENTATION_HORIZONTAL = 1;
    public static final int SCROLL_ORIENTATION_ISOMETRIC = 2;

// Local Constants -------------------------------------------------------------------------/

    private static final String APPTITLE = "Magellan v" + VERSION_NUMBER + " : TI-99/4A Map Editor";

    private final String FILEEXT = "mag";
    private final String[] FILEEXTS = {FILEEXT};
    private final String XBEXT = "xb";
    private final String[] XBEXTS = {XBEXT, "bas", "txt"};
    private final String ASMEXT = "asm";
    private final String[] ASMEXTS = {ASMEXT, "a99"};
    private final String IMGEXT = "png";
    private final String[] IMGEXTS = {IMGEXT, "gif"};
    private final String BINEXT = "mgb";
    private final String[] BINEXTS = {BINEXT};
    private final String VDPEXT = "vdp";
    private final String[] VDPEXTS = {VDPEXT, "vram", "bin"};
    private final String ANY = "";
    private final String[] ANYS = {ANY};

    private final int FONT_ROWS = 32;
    private final int BAS_FONT_ROWS = 16;
    private final int ADV_FONT_ROWS = 32;
    private final int FONT_COLS = 8;

    private final int MAP_ROWS = 24;
    private final int MAP_COLS = 32;
    private final int MAP_CELL = 8;

    private final Color CLR_CHARS_BASE1 = new Color(232, 232, 232);
    private final Color CLR_CHARS_BASE2 = new Color(196, 196, 196);
    private final Color CLR_CHARS_LOWER = new Color(222, 242, 255);
    private final Color CLR_CHARS_UPPER = new Color(255, 222, 242);

// Variables -------------------------------------------------------------------------------/

    protected boolean advancedModeOn = false;
    protected int colorMode = COLOR_MODE_GRAPHICS_1;

    protected HashMap<Integer, int[][]> hmCharGrids;
    protected HashMap<Integer, int[][]> hmCharColors;
    protected HashMap<Integer, int[][]> hmDefaultChars;
    protected HashMap<Integer, Image> hmCharImages;

    protected int[][] clrSets = new int[FONT_ROWS][2];
    protected ECMPalette[] ecmPalettes = null;
    protected ECMPalette[] ecmCharPalettes = null;
    boolean[] ecmCharTransparency = new boolean[256];

    protected int activeChar = TIGlobals.CUSTOMCHAR;
    protected int lastActiveChar = MapCanvas.NOCHAR;

    protected File mapDataFile;

    // Import / export settings

    protected boolean bExportComments = true;
    protected boolean bIncludeCharNumbers = true;
    protected boolean bCurrentMapOnly = false;
    protected boolean bSwapBoth = true;
    protected boolean bSwapImgs = true;
    protected boolean bAllMaps = true;

    protected int defStartChar = TIGlobals.FIRSTCHAR;
    protected int defEndChar = TIGlobals.LASTCHAR;

// Components ------------------------------------------------------------------------------/

    private JPanel jpnlMain;

    private JPanel jpnlEdit;

    private JPanel jpnlChar;
    private JPanel jpnlChTools;
    private JPanel jpnlColorDock;
    private GridCanvas gcChar;
    private JTextField jtxtChar;
    private JButton jbtnUpdateChar;
    private JButton jbtnLook;
    private JButton jbtnUndo;
    private JButton jbtnRedo;
    private JLabel jlblCharInt;
    private JLabel jlblCharHex;
    private JCheckBox jchkTransparency;

    private JPanel jpnlCharacterEditor;

    private JButton[] jbtnChar;
    private JPanel jpnlFont;
    private JScrollPane jsclFont;

    JMenuItem jmitGraphicsColorMode;
    JMenuItem jmitBitmapColorMode;
    JMenuItem jmitECM2ColorMode;
    JMenuItem jmitECM3ColorMode;

    private MapEditor mapdMain;

    private DualClickButton[] colorDockButtons = new DualClickButton[0];
    private ECMPaletteComboBox ecmPaletteComboBox;

    AnalyzeCharTransDialog charTransDialog;

    private Properties appProperties = new Properties();
    private String filePath;

// Constructors ----------------------------------------------------------------------------/

    public Magellan(String filePath) {
        super(APPTITLE);
        this.filePath = filePath;
    }

    public void run() {

        try {
            setIconImage(new ImageIcon(getClass().getResource("images/icon64.png")).getImage());
        } catch (Exception e) {
            e.printStackTrace();
        }

        // obtain application properties (if exist)
        try {
            File prefsFile = new File(System.getProperty("user.home") + "/Magellan.prefs");
            if (prefsFile.exists()) {
                FileInputStream fis = new FileInputStream(prefsFile);
                appProperties.load(fis);
                fis.close();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

        hmDefaultChars = new HashMap<Integer, int[][]>();
        for (int ch = TIGlobals.CHARMAPSTART; ch <= TIGlobals.CHARMAPEND; ch++) {
            hmDefaultChars.put(ch, Globals.getByteIntArray(TIGlobals.DEFAULT_TI_CHARS[ch - TIGlobals.CHARMAPSTART], 8));
        }

        if (appProperties.getProperty("exportComments") != null) {
            bExportComments = appProperties.getProperty("exportComments").toLowerCase().equals("true");
        }
        if (appProperties.getProperty("includeCharNumbers") != null) {
            bIncludeCharNumbers = appProperties.getProperty("includeCharNumbers").toLowerCase().equals("true");
        }
        if (appProperties.getProperty("currentMapOnly") != null) {
            bCurrentMapOnly = appProperties.getProperty("currentMapOnly").toLowerCase().equals("true");
        }
        if (appProperties.getProperty("expandCharacters") != null) {
            advancedModeOn = appProperties.getProperty("expandCharacters").toLowerCase().equals("true");
        }
        if (appProperties.getProperty("colorMode") != null) {
            colorMode = Integer.parseInt(appProperties.getProperty("colorMode"));
        }
        if (appProperties.getProperty("defStartChar") != null) {
            defStartChar = Integer.parseInt(appProperties.getProperty("defStartChar"));
        }
        if (appProperties.getProperty("defEndChar") != null) {
            defEndChar = Integer.parseInt(appProperties.getProperty("defEndChar"));
        }
        if (appProperties.getProperty("filePath") != null) {
            mapDataFile = new File(appProperties.getProperty("filePath"));
            if (mapDataFile != null && !mapDataFile.isDirectory()) {
                mapDataFile = mapDataFile.getParentFile();
            }
        }

        // Create map editor panel (needs to initialise early for the listeners)
        mapdMain = new MapEditor(MAP_COLS, MAP_ROWS, MAP_CELL, this, this);
        mapdMain.fillGrid(TIGlobals.SPACECHAR);
        mapdMain.setBkgrndColor(Globals.CLR_COMPONENTBACK);
        if (appProperties.getProperty("magnif") != null) {
            mapdMain.setViewScale(Integer.parseInt(appProperties.getProperty("magnif")));
        }
        if (appProperties.getProperty("textCursor") != null) {
            mapdMain.setTypeCellOn(appProperties.getProperty("textCursor").toLowerCase().equals("true"));
        }
        if (appProperties.getProperty("showGrid") != null) {
            mapdMain.setShowGrid(appProperties.getProperty("showGrid").toLowerCase().equals("true"));
        }
        if (appProperties.getProperty("showPosition") != null) {
            mapdMain.setShowPosIndic(appProperties.getProperty("showPosition").toLowerCase().equals("true"));
        }
        if (appProperties.getProperty("base0Position") != null) {
            mapdMain.setBase0Position(appProperties.getProperty("base0Position").toLowerCase().equals("true"));
        }
        mapdMain.addScreenColorListener(this);

        // Create the menus
        JMenuBar jMenuBar = new JMenuBar();
        JMenu jmenFile = new JMenu("File");
        JMenuItem jmitNew = new JMenuItem("New Map Project");
        jmitNew.setActionCommand(Globals.CMD_NEW);
        jmitNew.addActionListener(this);
        jmenFile.add(jmitNew);
        JMenuItem jmitOpen = new JMenuItem("Open Map Project");
        jmitOpen.setActionCommand(Globals.CMD_OPEN);
        jmitOpen.addActionListener(this);
        jmenFile.add(jmitOpen);
        JMenuItem jmitSave = new JMenuItem("Save Map Project");
        jmitSave.setActionCommand(Globals.CMD_SAVE);
        jmitSave.addActionListener(this);
        jmenFile.add(jmitSave);
        JMenuItem jmitSaveAs = new JMenuItem("Save Map Project as...");
        jmitSaveAs.setActionCommand(Globals.CMD_SAVEAS);
        jmitSaveAs.addActionListener(this);
        jmenFile.add(jmitSaveAs);
        jmenFile.addSeparator();
        JMenuItem jmitAppend = new JMenuItem("Append Maps");
        jmitAppend.setActionCommand(Globals.CMD_APPEND);
        jmitAppend.addActionListener(this);
        jmenFile.add(jmitAppend);
        jmenFile.addSeparator();
//		JMenuItem jmitLoadDefs = new JMenuItem("Load Default Characters"); jmitLoadDefs.setActionCommand(CMD_LOADDEFS); jmitLoadDefs.addActionListener(this); jmenFile.add(jmitLoadDefs);
//		jmenFile.addSeparator();
        JMenuItem jmitExit = new JMenuItem("Exit");
        jmitExit.setActionCommand(Globals.CMD_EXIT);
        jmitExit.addActionListener(this);
        jmenFile.add(jmitExit);
        jMenuBar.add(jmenFile);

        JMenu jmenImport = new JMenu("Import");
        JMenuItem jmitImportChrImgMono = new JMenuItem("Character Image (Mono)");
        jmitImportChrImgMono.setActionCommand(Globals.CMD_MPCIMGMN);
        jmitImportChrImgMono.addActionListener(this);
        jmenImport.add(jmitImportChrImgMono);
        JMenuItem jmitImportChrImgColor = new JMenuItem("Character Image (Color)");
        jmitImportChrImgColor.setActionCommand(Globals.CMD_MPCIMGCL);
        jmitImportChrImgColor.addActionListener(this);
        jmenImport.add(jmitImportChrImgColor);
        JMenuItem jmitImportVramDump = new JMenuItem("VRAM Dump");
        jmitImportVramDump.setActionCommand(Globals.CMD_MPVDP);
        jmitImportVramDump.addActionListener(this);
        jmenImport.add(jmitImportVramDump);
        JMenuItem jmitImportMapImage = new JMenuItem("Map Image");
        jmitImportMapImage.setActionCommand(Globals.CMD_MPMAPIMG);
        jmitImportMapImage.addActionListener(this);
        jmenImport.add(jmitImportMapImage);
        jMenuBar.add(jmenImport);

        JMenu jmenExport = new JMenu("Export");
        JMenuItem jmitExportData = new JMenuItem("BASIC Data");
        jmitExportData.setActionCommand(Globals.CMD_XPDATA);
        jmitExportData.addActionListener(this);
        jmenExport.add(jmitExportData);
        JMenuItem jmitExportExec = new JMenuItem("XB Program");
        jmitExportExec.setActionCommand(Globals.CMD_XPEXEC);
        jmitExportExec.addActionListener(this);
        jmenExport.add(jmitExportExec);
        JMenuItem jmitExportAsm = new JMenuItem("Assembler Data");
        jmitExportAsm.setActionCommand(Globals.CMD_XPASM);
        jmitExportAsm.addActionListener(this);
        jmenExport.add(jmitExportAsm);
        JMenuItem jmitExportScrollMap = new JMenuItem("Assembler Character Transition Data");
        jmitExportScrollMap.setActionCommand(Globals.CMD_XPSCROLL);
        jmitExportScrollMap.addActionListener(this);
        jmenExport.add(jmitExportScrollMap);
        JMenuItem jmitExportBin = new JMenuItem("Binary Data");
        jmitExportBin.setActionCommand(Globals.CMD_XPBIN);
        jmitExportBin.addActionListener(this);
        jmenExport.add(jmitExportBin);
        JMenuItem jmitExportBinMap = new JMenuItem("Binary Map (Current)");
        jmitExportBinMap.setActionCommand(Globals.CMD_XPBINMAP);
        jmitExportBinMap.addActionListener(this);
        jmenExport.add(jmitExportBinMap);
        JMenuItem jmitExportXBDisMer = new JMenuItem("XB Display Merge");
        jmitExportXBDisMer.setActionCommand(Globals.CMD_XPXBDISMER);
        jmitExportXBDisMer.addActionListener(this);
        jmenExport.add(jmitExportXBDisMer);
        jmenExport.addSeparator();
        JMenuItem jmitExportChrImgMono = new JMenuItem("Character Image (Mono)");
        jmitExportChrImgMono.setActionCommand(Globals.CMD_XPCIMGMN);
        jmitExportChrImgMono.addActionListener(this);
        jmenExport.add(jmitExportChrImgMono);
        JMenuItem jmitExportChrImgColor = new JMenuItem("Character Image (Color)");
        jmitExportChrImgColor.setActionCommand(Globals.CMD_XPCIMGCL);
        jmitExportChrImgColor.addActionListener(this);
        jmenExport.add(jmitExportChrImgColor);
        JMenuItem jmitExportMapImg = new JMenuItem("Map Image");
        jmitExportMapImg.setActionCommand(Globals.CMD_XPMAPIMG);
        jmitExportMapImg.addActionListener(this);
        jmenExport.add(jmitExportMapImg);
        jMenuBar.add(jmenExport);

        JMenu jmenTools = new JMenu("Tools");
        JMenuItem jmitSwapChars = new JMenuItem("Replace Characters");
        jmitSwapChars.setActionCommand(Globals.CMD_SWAPCHARS);
        jmitSwapChars.addActionListener(this);
        jmenTools.add(jmitSwapChars);
        JMenuItem jmitAnalyzeCharTrans = new JMenuItem("Analyze Character Transitions");
        jmitAnalyzeCharTrans.setActionCommand(Globals.CMD_ANALYZECHARTRANS);
        jmitAnalyzeCharTrans.addActionListener(this);
        jmenTools.add(jmitAnalyzeCharTrans);

        if (ISOMETRIC) {
            JMenuItem jmitAnalyzeIsoCharTrans = new JMenuItem();
            jmitAnalyzeIsoCharTrans.setAction(
                new AbstractAction() {
                    {
                        putValue(Action.NAME, "Analyze Isometric Character Transitions");
                    }
                    public void actionPerformed(ActionEvent e) {
                        Map<String, Integer> transMap = new HashMap<String, Integer>();
                        int[][] mapData = mapdMain.getMapData(mapdMain.getCurrentMapId());
                        for (int y = 0; y < mapData.length - 1; y++) {
                            for (int x = 0; x < mapData[y].length - 2; x++) {
                                String key =
                                    mapData[y][x] + "-" + mapData[y][x + 1] + "-" + mapData[y][x + 2] + "-" +
                                    mapData[y + 1][x] + "-" + mapData[y + 1][x + 1] + "-" + mapData[y + 1][x + 2];
                                Integer count = transMap.get(key);
                                transMap.put(key, count == null ? 1 : count + 1);
                            }
                        }
                        for (String key : transMap.keySet()) {
                            System.out.println(key + ": " + transMap.get(key));
                        }
                        informationAction(Magellan.this, "Analyze Isometric Character Transitions", "Total transitions: " + Integer.toString(transMap.size()));
                    }
                }
            );
            jmenTools.add(jmitAnalyzeIsoCharTrans);
        }

        jMenuBar.add(jmenTools);

        JMenu jmenOptions = new JMenu("Options");
        JMenuItem jmitShowPos = new JCheckBoxMenuItem("Show Position", mapdMain.showPosIndic());
        jmitShowPos.setActionCommand(Globals.CMD_SHOWPOS);
        jmitShowPos.addActionListener(this);
        jmenOptions.add(jmitShowPos);
        JMenuItem jmitBase0Pos = new JCheckBoxMenuItem("Base 0 for Position", mapdMain.base0Position());
        jmitBase0Pos.setActionCommand(Globals.CMD_BASE0POS);
        jmitBase0Pos.addActionListener(this);
        jmenOptions.add(jmitBase0Pos);
        JMenuItem jmitChangeMode = new JCheckBoxMenuItem("Expanded Characters", advancedModeOn);
        jmitChangeMode.setActionCommand(Globals.CMD_MODESET);
        jmitChangeMode.addActionListener(this);
        jmenOptions.add(jmitChangeMode);
        jmenOptions.addSeparator();
        ButtonGroup colorModeButtonGroup = new ButtonGroup();
        jmitGraphicsColorMode = new JRadioButtonMenuItem(COLOR_MODES[COLOR_MODE_GRAPHICS_1], colorMode == COLOR_MODE_GRAPHICS_1);
        colorModeButtonGroup.add(jmitGraphicsColorMode);
        jmitGraphicsColorMode.setActionCommand(Globals.CMD_GRAPHICSCOLORMODE);
        jmitGraphicsColorMode.addActionListener(this);
        jmenOptions.add(jmitGraphicsColorMode);
        jmitBitmapColorMode = new JRadioButtonMenuItem(COLOR_MODES[COLOR_MODE_BITMAP], colorMode == COLOR_MODE_BITMAP);
        colorModeButtonGroup.add(jmitBitmapColorMode);
        jmitBitmapColorMode.setActionCommand(Globals.CMD_BITMAPCOLORMODE);
        jmitBitmapColorMode.addActionListener(this);
        jmenOptions.add(jmitBitmapColorMode);
        jmitECM2ColorMode = new JRadioButtonMenuItem(COLOR_MODES[COLOR_MODE_ECM_2], colorMode == COLOR_MODE_ECM_2);
        colorModeButtonGroup.add(jmitECM2ColorMode);
        jmitECM2ColorMode.setActionCommand(Globals.CMD_ECM2COLORMODE);
        jmitECM2ColorMode.addActionListener(this);
        jmenOptions.add(jmitECM2ColorMode);
        jmitECM3ColorMode = new JRadioButtonMenuItem(COLOR_MODES[COLOR_MODE_ECM_3], colorMode == COLOR_MODE_ECM_3);
        colorModeButtonGroup.add(jmitECM3ColorMode);
        jmitECM3ColorMode.setActionCommand(Globals.CMD_ECM3COLORMODE);
        jmitECM3ColorMode.addActionListener(this);
        jmenOptions.add(jmitECM3ColorMode);
        jMenuBar.add(jmenOptions);

        JMenu jmenHelp = new JMenu("Help");
        JMenuItem jmitHelpAbout = new JMenuItem("About Magellan");
        jmitHelpAbout.setActionCommand(Globals.CMD_ABOUT);
        jmitHelpAbout.addActionListener(this);
        jmenHelp.add(jmitHelpAbout);
        jMenuBar.add(jmenHelp);

        // Create the main panel
        jpnlMain = getPanel(new BorderLayout());

        // Create Character Editor
        jpnlChTools = getPanel(new GridBagLayout());
        GridBagConstraints gbci = new GridBagConstraints();
        gbci.anchor = GridBagConstraints.CENTER;
        gbci.fill = GridBagConstraints.NONE;
        gbci.gridheight = 1;
        gbci.gridwidth = 1;
        gbci.gridx = 1;
        gbci.gridy = 1;
        gbci.insets = new Insets(1, 1, 1, 1);
        gbci.ipadx = 2;
        gbci.ipady = 2;
        gbci.weightx = 1;
        gbci.weighty = 1;

        gbci.gridx = 1;
        gbci.gridy = 1;
        gbci.gridheight = 7;
        gbci.gridwidth = 1;
        gbci.anchor = GridBagConstraints.NORTH;
        JPanel jpnlToolButtons = getPanel(new GridLayout(8, 1, 0, 2));
        jpnlToolButtons.add(getToolButton(Globals.CMD_FILL, "Fill"));
        jpnlToolButtons.add(getToolButton(Globals.CMD_CLEAR, "Clear"));
        jpnlToolButtons.add(getToolButton(Globals.CMD_INVERT, "Invert Image"));
        jpnlToolButtons.add(getToolButton(Globals.CMD_GRID, "Toggle Grid"));
        jbtnLook = getToolButton(Globals.CMD_LOOK, "Look At Character");
        jpnlToolButtons.add(jbtnLook);
        jbtnUndo = getToolButton(Globals.CMD_UNDO, "Undo Edit");
        jbtnUndo.setEnabled(false);
        jpnlToolButtons.add(jbtnUndo);
        jbtnRedo = getToolButton(Globals.CMD_REDO, "Redo Edit");
        jbtnRedo.setEnabled(false);
        jpnlToolButtons.add(jbtnRedo);
        jchkTransparency = new JCheckBox();
        jchkTransparency.setOpaque(false);
        jchkTransparency.setToolTipText("Toggle Transparency");
        jchkTransparency.setVisible(colorMode == COLOR_MODE_ECM_2 || colorMode == COLOR_MODE_ECM_3);
        jchkTransparency.setActionCommand(Globals.CMD_TRANSPARENCY);
        jchkTransparency.addActionListener(this);
        jpnlToolButtons.add(jchkTransparency);
        jpnlChTools.add(jpnlToolButtons, gbci);

        gbci.gridx = 2;
        gbci.gridy = 1;
        gbci.gridheight = 1;
        gbci.gridwidth = 1;
        gbci.anchor = GridBagConstraints.CENTER;
        jpnlChTools.add(getToolButton(Globals.CMD_ROTATEL, "Rotate Left", Globals.CLR_BUTTON_TRANS), gbci);
        gbci.gridx = 3;
        gbci.gridy = 1;
        jlblCharInt = getLabel("", JLabel.CENTER);
        jlblCharInt.setPreferredSize(Globals.DM_TOOL);
        jpnlChTools.add(jlblCharInt, gbci);
        gbci.gridx = 4;
        gbci.gridy = 1;
        jpnlChTools.add(getToolButton(Globals.CMD_SHIFTU, "Shift Up", Globals.CLR_BUTTON_SHIFT), gbci);
        gbci.gridx = 5;
        gbci.gridy = 1;
        jlblCharHex = getLabel("", JLabel.CENTER);
        jlblCharHex.setPreferredSize(Globals.DM_TOOL);
        jpnlChTools.add(jlblCharHex, gbci);
        gbci.gridx = 6;
        gbci.gridy = 1;
        jpnlChTools.add(getToolButton(Globals.CMD_ROTATER, "Rotate Right", Globals.CLR_BUTTON_TRANS), gbci);
        gbci.gridx = 2;
        gbci.gridy = 3;
        jpnlChTools.add(getToolButton(Globals.CMD_SHIFTL, "Shift Left", Globals.CLR_BUTTON_SHIFT), gbci);
        gbci.gridx = 6;
        gbci.gridy = 3;
        jpnlChTools.add(getToolButton(Globals.CMD_SHIFTR, "Shift Right", Globals.CLR_BUTTON_SHIFT), gbci);
        gbci.gridx = 2;
        gbci.gridy = 5;
        jpnlChTools.add(getToolButton(Globals.CMD_FLIPH, "Flip Horizontal", Globals.CLR_BUTTON_TRANS), gbci);
        gbci.gridx = 4;
        gbci.gridy = 5;
        jpnlChTools.add(getToolButton(Globals.CMD_SHIFTD, "Shift Down", Globals.CLR_BUTTON_SHIFT), gbci);
        gbci.gridx = 6;
        gbci.gridy = 5;
        jpnlChTools.add(getToolButton(Globals.CMD_FLIPV, "Flip Vertical", Globals.CLR_BUTTON_TRANS), gbci);

        gbci.gridx = 3;
        gbci.gridy = 2;
        gbci.gridheight = 3;
        gbci.gridwidth = 3;
        gbci.fill = GridBagConstraints.BOTH;
        gbci.weightx = 2;
        gbci.weighty = 2;
        gbci.anchor = GridBagConstraints.CENTER;
        jpnlChar = getPanel(new BorderLayout());
        gcChar = new GridCanvas(TIGlobals.TI_PALETTE_OPAQUE, 8, 8, 8, this, this, colorMode);
        gcChar.setColorScreen(TIGlobals.TI_PALETTE_OPAQUE[mapdMain.getColorScreen()]);
        jpnlChar.add(gcChar, BorderLayout.CENTER);
        Dimension jpnlCharDimension = new Dimension(136, 136);
        jpnlChar.setPreferredSize(jpnlCharDimension);
        jpnlChar.setMinimumSize(jpnlCharDimension);
        jpnlChTools.add(jpnlChar, gbci);

        JPanel jpnlCharTool = getPanel(new BorderLayout());
        jtxtChar = new JTextField();
        jpnlCharTool.add(jtxtChar, BorderLayout.CENTER);
        JPanel jpnlCharToolbar = getPanel(new GridLayout(1, 3));
        jbtnUpdateChar = getToolButton(Globals.CMD_UPDATE, "Set Char");
        jbtnUpdateChar.addActionListener(this);
        jpnlCharToolbar.add(jbtnUpdateChar);
        JButton copyButton = getToolButton(Globals.CMD_COPY, "Copy");
        copyButton.addActionListener(this);
        jpnlCharToolbar.add(copyButton);
        JButton pasteButton = getToolButton(Globals.CMD_PASTE, "Paste and set");
        pasteButton.addActionListener(this);
        jpnlCharToolbar.add(pasteButton);
        jpnlCharTool.add(jpnlCharToolbar, BorderLayout.EAST);
        gbci.gridx = 1;
        gbci.gridy = 6;
        gbci.gridheight = 1;
        gbci.gridwidth = 6;
        gbci.fill = GridBagConstraints.HORIZONTAL;
        gbci.weightx = 1;
        gbci.weighty = 1;
        jpnlChTools.add(jpnlCharTool, gbci);

        // Create Color Dock
        buildColorDock();

        // Create Character Dock Buttons
        hmCharGrids = new HashMap<Integer, int[][]>();
        if (colorMode == COLOR_MODE_BITMAP) {
            hmCharColors = new HashMap<Integer, int[][]>();
        }
        hmCharImages = new HashMap<Integer, Image>();
        jbtnChar = new JButton[(TIGlobals.MAXCHAR - TIGlobals.MINCHAR) + 1];
        for (int ch = TIGlobals.MINCHAR; ch <= TIGlobals.MAXCHAR; ch++) {
            int rowNum = (int) (Math.floor(ch / 8));
            clrSets[rowNum][Globals.INDEX_CLR_BACK] = 0;
            clrSets[rowNum][Globals.INDEX_CLR_FORE] = 1;
            jbtnChar[ch] = getDockButton(((ch >= TIGlobals.CHARMAPSTART) && (ch <= TIGlobals.CHARMAPEND) ? "" + TIGlobals.CHARMAP[ch - TIGlobals.CHARMAPSTART] : "?"), Globals.CMD_EDITCHR + ch, TIGlobals.TI_PALETTE_OPAQUE[clrSets[rowNum][Globals.INDEX_CLR_BACK]]);
            jbtnChar[ch].setForeground(TIGlobals.TI_COLOR_UNUSED);
            int[][] emptyGrid = new int[8][8];
            for (int y = 0; y < emptyGrid.length; y++) {
                for (int x = 0; x < emptyGrid[y].length; x++) {
                    if (ch >= TIGlobals.CHARMAPSTART && ch <= TIGlobals.CHARMAPEND) {
                        emptyGrid[y][x] = hmDefaultChars.get(ch)[y][x];
                    } else {
                        emptyGrid[y][x] = 0;
                    }
                }
            }
            hmCharGrids.put(ch, emptyGrid);
            if (colorMode == COLOR_MODE_BITMAP) {
                int[][] emptyColors = new int[8][2];
                for (int y = 0; y < emptyColors.length; y++) {
                    emptyColors[y][0] = 0;
                    emptyColors[y][1] = 1;
                }
                hmCharColors.put(ch, emptyColors);
            }
        }

        // Create Character Dock
        buildCharacterDock();
        JPanel jpnlNochange = getPanel(new FlowLayout());
        jpnlNochange.add(jpnlFont);
        jsclFont = new JScrollPane(jpnlNochange, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

        // Assemble Editor Panel
        jpnlEdit = getPanel(new GridBagLayout());
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.anchor = GridBagConstraints.NORTH;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.insets = new Insets(4, 0, 0, 0);
        gbc.weightx = 1;
        gbc.weighty = 0;
        gbc.gridx = 1;
        gbc.gridy = 1;
        jpnlEdit.add(jpnlChTools, gbc);
        gbc.gridx = 1;
        gbc.gridy = 2;
        jpnlEdit.add(jpnlColorDock, gbc);
        gbc.gridx = 1;
        gbc.gridy = 3;
        gbc.weightx = 1;
        gbc.weighty = 1;
        gbc.fill = GridBagConstraints.BOTH;
        jpnlEdit.add(jsclFont, gbc);

        jpnlMain.add(mapdMain, BorderLayout.CENTER);
        jpnlMain.add(jpnlEdit, BorderLayout.WEST);

        // Assemble the application
        this.getContentPane().setLayout(new BorderLayout());
        this.getContentPane().add(jpnlMain, BorderLayout.CENTER);
        this.getContentPane().add(jMenuBar, BorderLayout.NORTH);
        this.addWindowListener(this);
        this.addMouseListener(this);
        this.addMouseMotionListener(this);
        this.pack();
        this.setVisible(true);
        this.setExtendedState(JFrame.MAXIMIZED_BOTH);

        // Initialise the border objects for button selection
        Globals.bordButtonNormal = jbtnChar[0].getBorder();

        jpnlChTools.setMinimumSize(jpnlChTools.getPreferredSize());
        jpnlChTools.setPreferredSize(jpnlChTools.getPreferredSize());
        jpnlChTools.setSize(jpnlChTools.getPreferredSize());

        // jpnlColorDock.setMinimumSize(jpnlColorDock.getPreferredSize());
        // jpnlColorDock.setPreferredSize(jpnlColorDock.getPreferredSize());
        // jpnlColorDock.setSize(jpnlColorDock.getPreferredSize());

        updateFontButtons();
        gcChar.addUndoRedoListener(this);
        mapdMain.resetUndoManager();

        SwingUtilities.invokeLater(
            new Runnable() {
                public void run() {
                    activeChar = TIGlobals.CUSTOMCHAR;
                    ActionEvent aeInit = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, Globals.CMD_EDITCHR + activeChar);
                    Magellan.this.actionPerformed(aeInit);
                    if (filePath != null) {
                        openDataFile(filePath);
                    }
                }
            }
        );
    }

// Listeners -------------------------------------------------------------------------------/

    /* ActionListener methods */
    public void actionPerformed(ActionEvent ae) {
        try {
            String command = ae.getActionCommand();
            if (command.equals(Globals.CMD_EXIT)) {
                exitApp(0);
            } else if (command.equals(Globals.CMD_CLEAR)) {
                gcChar.clearGrid();
                hmCharGrids.put(activeChar, gcChar.getGridData());
                if (colorMode == COLOR_MODE_BITMAP) {
                    hmCharColors.put(activeChar, gcChar.getGridColors());
                }
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_FILL)) {
                gcChar.fillGrid();
                hmCharGrids.put(activeChar, gcChar.getGridData());
                if (colorMode == COLOR_MODE_BITMAP) {
                    hmCharColors.put(activeChar, gcChar.getGridColors());
                }
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_GRID)) {
                gcChar.toggleGrid();
                updateComponents();
            } else if (command.equals(Globals.CMD_LOOK)) {
                mapdMain.setLookModeOn(!mapdMain.isLookModeOn());
                jbtnLook.setBackground((mapdMain.isLookModeOn() ? Globals.CLR_BUTTON_ACTIVE : Globals.CLR_BUTTON_NORMAL));
            } else if (command.equals(Globals.CMD_UNDO)) {
                gcChar.undo();
            } else if (command.equals(Globals.CMD_REDO)) {
                gcChar.redo();
            } else if (command.equals(Globals.CMD_FLIPH)) {
                gcChar.setGrid(Globals.flipGrid(gcChar.getGridData(), false));
                hmCharGrids.put(activeChar, gcChar.getGridData());
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_FLIPV)) {
                gcChar.setGridAndColors(Globals.flipGrid(gcChar.getGridData(), true), colorMode == COLOR_MODE_BITMAP ? Globals.flipGrid(gcChar.getGridColors(), true) : null);
                hmCharGrids.put(activeChar, gcChar.getGridData());
                if (colorMode == COLOR_MODE_BITMAP) {
                    hmCharColors.put(activeChar, gcChar.getGridColors());
                }
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_ROTATEL)) {
                gcChar.setGrid(Globals.rotateGrid(gcChar.getGridData(), true));
                hmCharGrids.put(activeChar, gcChar.getGridData());
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_ROTATER)) {
                gcChar.setGrid(Globals.rotateGrid(gcChar.getGridData(), false));
                hmCharGrids.put(activeChar, gcChar.getGridData());
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_INVERT)) {
                gcChar.setGrid(Globals.invertGrid(gcChar.getGridData(), colorMode == COLOR_MODE_GRAPHICS_1 || colorMode == COLOR_MODE_BITMAP ? 1 : (colorMode == COLOR_MODE_ECM_2 ? 3 : 7)));
                hmCharGrids.put(activeChar, gcChar.getGridData());
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_SHIFTU)) {
                gcChar.setGrid(Globals.cycleGridUp(gcChar.getGridData()));
                hmCharGrids.put(activeChar, gcChar.getGridData());
                if (colorMode == COLOR_MODE_BITMAP) {
                    gcChar.setColors(Globals.cycleGridUp(gcChar.getGridColors()));
                    hmCharColors.put(activeChar, gcChar.getGridColors());
                }
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_SHIFTD)) {
                gcChar.setGrid(Globals.cycleGridDown(gcChar.getGridData()));
                hmCharGrids.put(activeChar, gcChar.getGridData());
                if (colorMode == COLOR_MODE_BITMAP) {
                    gcChar.setColors(Globals.cycleGridDown(gcChar.getGridColors()));
                    hmCharColors.put(activeChar, gcChar.getGridColors());
                }
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_SHIFTL)) {
                gcChar.setGrid(Globals.cycleGridLeft(gcChar.getGridData()));
                hmCharGrids.put(activeChar, gcChar.getGridData());
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_SHIFTR)) {
                gcChar.setGrid(Globals.cycleGridRight(gcChar.getGridData()));
                hmCharGrids.put(activeChar, gcChar.getGridData());
                updateFontButton(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_NEW)) {
                int userResponse = confirmationAction(this, "Confirm New Project", "This will delete all current data.\n\rAre you sure?", false);
                if (userResponse == JOptionPane.YES_OPTION) {
                    newProject();
                    this.setTitle(APPTITLE);
                    activeChar = TIGlobals.CUSTOMCHAR;
                    ActionEvent aeInit = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, Globals.CMD_EDITCHR + activeChar);
                    this.actionPerformed(aeInit);
                }
            } else if (command.equals(Globals.CMD_OPEN)) {
                openDataFile(null);
            } else if (command.equals(Globals.CMD_SAVE)) {
                saveDataFile();
            } else if (command.equals(Globals.CMD_SAVEAS)) {
                saveDataFileAs();
            } else if (command.equals(Globals.CMD_APPEND)) {
                appendDataFile();
            } else if (command.equals(Globals.CMD_LOADDEFS)) {
                loadDefaultCharacters();
            } else if (command.equals(Globals.CMD_MPCIMGMN)) {
                importCharImage(false);
            } else if (command.equals(Globals.CMD_MPCIMGCL)) {
                importCharImage(true);
            } else if (command.equals(Globals.CMD_MPVDP)) {
                importVramDump();
            } else if (command.equals(Globals.CMD_MPMAPIMG)) {
                importMapImage();
            } else if (command.equals(Globals.CMD_XPDATA)) {
                exportDataFile(false);
            } else if (command.equals(Globals.CMD_XPEXEC)) {
                exportDataFile(true);
            } else if (command.equals(Globals.CMD_XPASM)) {
                exportAssemblerFile();
            } else if (command.equals(Globals.CMD_XPSCROLL)) {
                exportScrollFile();
            } else if (command.equals(Globals.CMD_XPBIN)) {
                exportBinaryFile();
            } else if (command.equals(Globals.CMD_XPBINMAP)) {
                exportBinaryMapFile();
            } else if (command.equals(Globals.CMD_XPXBDISMER)) {
                exportXBDisplayMerge();
            } else if (command.equals(Globals.CMD_XPCIMGMN)) {
                exportCharImage(false);
            } else if (command.equals(Globals.CMD_XPCIMGCL)) {
                exportCharImage(true);
            } else if (command.equals(Globals.CMD_XPMAPIMG)) {
                exportMapImage();
            } else if (command.startsWith(Globals.CMD_EDITCHR)) {
                activeChar = Integer.parseInt(command.substring(Globals.CMD_EDITCHR.length()));
                if (hmCharGrids.get(activeChar) == null) {
                    gcChar.clearGrid();
                    hmCharGrids.put(activeChar, gcChar.getGridData());
                    if (colorMode == COLOR_MODE_BITMAP) {
                        hmCharColors.put(activeChar, gcChar.getGridColors());
                    }
                }
                gcChar.resetUndoRedo();
                gcChar.setGridAndColors(hmCharGrids.get(activeChar), colorMode == COLOR_MODE_BITMAP ? hmCharColors.get(activeChar) : null);
                gcChar.setGridAndColors(hmCharGrids.get(activeChar), colorMode == COLOR_MODE_BITMAP ? hmCharColors.get(activeChar) : null);
                if (colorMode == COLOR_MODE_GRAPHICS_1) {
                    int cset = (int) (Math.floor(activeChar / 8));
                    gcChar.setColorBack(clrSets[cset][Globals.INDEX_CLR_BACK]);
                    gcChar.setColorDraw(clrSets[cset][Globals.INDEX_CLR_FORE]);
                    for (int i = 0; i < colorDockButtons.length; i++) {
                        colorDockButtons[i].setText(i == gcChar.getColorBack() ? "B" : (i == gcChar.getColorDraw() ? "F" : ""));
                    }
                }
                else if (colorMode == COLOR_MODE_ECM_2 || colorMode == COLOR_MODE_ECM_3) {
                    gcChar.setPalette(ecmCharPalettes[activeChar].getColors());
                    ecmPaletteComboBox.setSelectedItem(ecmCharPalettes[activeChar]);
                }
                gcChar.setECMTranparency(ecmCharTransparency[activeChar]);
                gcChar.redrawCanvas();
                mapdMain.setActiveChar(activeChar);
                updateComponents();
            } else if (command.equals(Globals.CMD_UPDATE)) {
                String hexString = null;
                switch (colorMode) {
                    case COLOR_MODE_GRAPHICS_1:
                        hexString = Globals.padHexString(jtxtChar.getText(), 16);
                        break;
                    case COLOR_MODE_BITMAP:
                        hexString = Globals.padHexString(jtxtChar.getText(), jtxtChar.getText().length() <= 16 ? 16 : 32);
                        break;
                    case COLOR_MODE_ECM_2:
                        hexString = Globals.padHexString(jtxtChar.getText(), 32);
                        break;
                    case COLOR_MODE_ECM_3:
                        hexString = Globals.padHexString(jtxtChar.getText(), 48);
                        break;
                }
                jtxtChar.setText(hexString);
                jtxtChar.setCaretPosition(0);
                // Plane 0
                int[][] charGrid = Globals.getByteIntArray(hexString.substring(0, 16), 8);
                // Bitmap colors
                int[][] charColors = hmCharColors.get(activeChar);
                if (colorMode == COLOR_MODE_BITMAP && hexString.length() == 32) {
                    charColors = Globals.parseColorHexString(hexString.substring(16));
                    hmCharColors.put(activeChar, charColors);
                }
                // Plane 1
                if (colorMode == COLOR_MODE_ECM_2 || colorMode == COLOR_MODE_ECM_3) {
                    Globals.orGrid(Globals.getByteIntArray(hexString.substring(16, 32), 8), charGrid, 1);
                }
                // Plane 2
                if (colorMode == COLOR_MODE_ECM_3) {
                    Globals.orGrid(Globals.getByteIntArray(hexString.substring(32, 48), 8), charGrid, 2);
                }
                hmCharGrids.put(activeChar, charGrid);
                gcChar.setGridAndColors(hmCharGrids.get(activeChar), colorMode == COLOR_MODE_BITMAP ? charColors : null);
                updateFontButton(activeChar);
            } else if (command.equals(Globals.CMD_COPY)) {
                jtxtChar.selectAll();
                jtxtChar.copy();
            } else if (command.equals(Globals.CMD_PASTE)) {
                jtxtChar.selectAll();
                jtxtChar.paste();
                jbtnUpdateChar.doClick();
            } else if (command.startsWith(Globals.CMD_CLRFORE)) {
                int clrval = Integer.parseInt(command.substring(Globals.CMD_CLRFORE.length()));
                if (colorMode == COLOR_MODE_GRAPHICS_1 || colorMode == COLOR_MODE_BITMAP) {
                    // Mark the selected foreground color
                    for (DualClickButton colorButton : colorDockButtons) {
                        if ("F".equals(colorButton.getText())) {
                            colorButton.setText("");
                        }
                    }
                    colorDockButtons[clrval].setText("F");
                    if (colorMode != COLOR_MODE_BITMAP) {
                        int cset = (int) (Math.floor(activeChar / 8));
                        clrSets[cset][Globals.INDEX_CLR_FORE] = clrval;
                        for (int c = 0; c < FONT_COLS; c++) {
                            updateFontButton((cset * 8) + c, false);
                        }
                        updateFontButton(activeChar);
                    }
                    gcChar.setColorDraw(clrval);
                    gcChar.redrawCanvas();
                }
                else {
                    gcChar.setColorDraw(clrval);
                }
            } else if (command.startsWith(Globals.CMD_CLRBACK)) {
                int clrval = Integer.parseInt(command.substring(Globals.CMD_CLRBACK.length()));
                if (colorMode == COLOR_MODE_GRAPHICS_1 || colorMode == COLOR_MODE_BITMAP) {
                        // Mark the selected background color
                    for (DualClickButton colorButton : colorDockButtons) {
                        if ("B".equals(colorButton.getText())) {
                            colorButton.setText("");
                        }
                    }
                    colorDockButtons[clrval].setText("B");
                    if (colorMode != COLOR_MODE_BITMAP) {
                        int cset = (int) (Math.floor(activeChar / 8));
                        clrSets[cset][Globals.INDEX_CLR_BACK] = clrval;
                        for (int c = 0; c < FONT_COLS; c++) {
                            updateFontButton((cset * 8) + c, false);
                        }
                        updateFontButton(activeChar);
                    }
                    gcChar.setColorBack(clrval);
                    gcChar.redrawCanvas();
                }
                else {
                    gcChar.setColorBack(clrval);
                }
            } else if (command.equals(Globals.CMD_PALSELECT)) {
                ECMPalette ecmPalette = ecmPalettes[ecmPaletteComboBox.getSelectedIndex()];
                ecmCharPalettes[activeChar] = ecmPalette;
                gcChar.setPalette(ecmPalette.getColors());
                gcChar.redrawCanvas();
                updateFontButton(activeChar);
            } else if (command.startsWith(Globals.CMD_CLRCHOOSE)) {
                int clrval = Integer.parseInt(command.substring(Globals.CMD_CLRCHOOSE.length()));
                ECMPalette ecmPalette = ecmPalettes[ecmPaletteComboBox.getSelectedIndex()];
                Color color = ECMColorChooser.showDialog(this, "Select Color", ecmPalette.getColor(clrval));
                if (color != null) {
                    ecmPaletteComboBox.setEditable(false);
                    ecmPalette.setColor(clrval, color);
                    ecmPaletteComboBox.setEditable(true);
                    gcChar.setPalette(ecmPalette.getColors());
                    gcChar.setColorScreen(getScreenColorPalette()[mapdMain.getColorScreen()]);
                    gcChar.redrawCanvas();
                    updateFontButtons();
                    mapdMain.setScreenColorPalette(getScreenColorPalette());
                }
            } else if (command.equals(Globals.CMD_SWAPCHARS)) {
                swapCharacters();
            } else if (command.equals(Globals.CMD_ANALYZECHARTRANS)) {
                analyzeCharTrans();
            }
            else if (command.equals(Globals.CMD_SHOWPOS)) {
                mapdMain.toggleShowPosIndic();
            } else if (command.equals(Globals.CMD_BASE0POS)) {
                mapdMain.toggleBase0Position();
            } else if (command.equals(Globals.CMD_CHARTOOL)) {
                jpnlCharacterEditor.setVisible(!jpnlCharacterEditor.isVisible());
            } else if (command.equals(Globals.CMD_MODESET)) {
                advancedModeOn = !advancedModeOn;
                buildCharacterDock();
            } else if (command.equals(Globals.CMD_GRAPHICSCOLORMODE)) {
                if (colorMode == COLOR_MODE_ECM_2 || colorMode == COLOR_MODE_ECM_3) {
                    limitCharGrids(2, colorMode == COLOR_MODE_ECM_2 ? 4 : 8);
                }
                colorMode = COLOR_MODE_GRAPHICS_1;
                buildColorDock();
                gcChar.setColorMode(colorMode, TIGlobals.TI_PALETTE_OPAQUE);
                int[] clrSet = clrSets[activeChar / 8];
                gcChar.setColorBack(clrSet[Globals.INDEX_CLR_BACK]);
                gcChar.setColorDraw(clrSet[Globals.INDEX_CLR_FORE]);
                gcChar.redrawCanvas();
                for (int i = 0; i < colorDockButtons.length; i++) {
                    colorDockButtons[i].setText(i == gcChar.getColorBack() ? "B" : (i == gcChar.getColorDraw() ? "F" : ""));
                }
                updateFontButtons();
            } else if (command.equals(Globals.CMD_BITMAPCOLORMODE)) {
                hmCharColors = new HashMap<Integer, int[][]>();
                for (int ch = TIGlobals.MINCHAR; ch <= TIGlobals.MAXCHAR; ch++) {
                    int[][] emptyColors = new int[8][2];
                    for (int y = 0; y < emptyColors.length; y++) {
                        int[] clrSet = clrSets[ch / 8];
                        if (colorMode == COLOR_MODE_GRAPHICS_1) {
                            emptyColors[y][Globals.INDEX_CLR_BACK] = clrSet[Globals.INDEX_CLR_BACK];
                            emptyColors[y][Globals.INDEX_CLR_FORE] = clrSet[Globals.INDEX_CLR_FORE];
                        }
                        else {
                            int[][] charGrid = hmCharGrids.get(ch);
                            if (charGrid != null) {
                                int[] row = charGrid[y];
                                ECMPalette ecmPalette = ecmCharPalettes[ch];
                                int[][] palette = new int[8][2];
                                int maxIndex = -1;
                                for (Integer colorIndex : row) {
                                    boolean found = false;
                                    for (int p = 0; p <= maxIndex && !found; p++) {
                                        if (palette[p][0] == colorIndex) {
                                            palette[p][1]++;
                                            found = true;
                                        }
                                    }
                                    if (!found) {
                                        maxIndex++;
                                        palette[maxIndex][0] = colorIndex;
                                        palette[maxIndex][1] = 1;
                                    }
                                }
                                Globals.sortGrid(palette);
                                emptyColors[y][Globals.INDEX_CLR_BACK] = Globals.getClosestColorIndex(ecmPalette.getColor(palette[0][0]), TIGlobals.TI_PALETTE_OPAQUE);
                                emptyColors[y][Globals.INDEX_CLR_FORE] = maxIndex > 0  ? Globals.getClosestColorIndex(ecmPalette.getColor(palette[1][0]), TIGlobals.TI_PALETTE_OPAQUE, emptyColors[y][Globals.INDEX_CLR_BACK]) : emptyColors[y][Globals.INDEX_CLR_BACK];
                                for (int x = 0; x < row.length; x++) {
                                    row[x] = Globals.colorDistance(ecmPalette.getColor(row[x]), TIGlobals.TI_PALETTE_OPAQUE[emptyColors[y][Globals.INDEX_CLR_BACK]]) <  Globals.colorDistance(ecmPalette.getColor(row[x]), TIGlobals.TI_PALETTE_OPAQUE[emptyColors[y][Globals.INDEX_CLR_FORE]]) ? 0 : 1;
                                }
                            }
                        }
                    }
                    hmCharColors.put(ch, emptyColors);
                }
                colorMode = COLOR_MODE_BITMAP;
                buildColorDock();
                gcChar.setColorMode(colorMode, TIGlobals.TI_PALETTE_OPAQUE);
                gcChar.redrawCanvas();
                for (int i = 0; i < colorDockButtons.length; i++) {
                    colorDockButtons[i].setText(i == gcChar.getColorBack() ? "B" : (i == gcChar.getColorDraw() ? "F" : ""));
                }
                updateFontButtons();
            } else if (command.equals(Globals.CMD_ECM2COLORMODE)) {
                if (colorMode == COLOR_MODE_ECM_3) {
                    limitCharGrids(4, 8);
                }
                colorMode = COLOR_MODE_ECM_2;
                buildColorDock();
                gcChar.setColorMode(colorMode, ecmCharPalettes[activeChar].getColors());
                updateFontButtons();
            } else if (command.equals(Globals.CMD_ECM3COLORMODE)) {
                colorMode = COLOR_MODE_ECM_3;
                buildColorDock();
                gcChar.setColorMode(colorMode, ecmCharPalettes[activeChar].getColors());
                updateFontButtons();
            } else if (command.equals(Globals.CMD_TRANSPARENCY)) {
                ecmCharTransparency[activeChar] = jchkTransparency.isSelected();
                gcChar.setECMTranparency(ecmCharTransparency[activeChar]);
                gcChar.redrawCanvas();
                updateFontButton(activeChar, true);
            } else if (command.equals(Globals.CMD_ABOUT)) {
                informationAction(
                    this,
                    "About Magellan",
                    "Magellan, version " + VERSION_NUMBER + ".\n\n" +
                    "\u00a9 2010 Howard Kistler/Dream Codex Retrogames\n" +
                    "www.dreamcodex.com\n\n" +
                    "Magellan is free software.\n\n" +
                    "Modified by retroclouds in 2011\n" +
                    "Modified by sometimes99er in 2013\n" +
                    "Modified by Rasmus Moustgaard 2013 -\n\n" +
                    "Java runtime version: " + System.getProperty("java.version")
                );
            }
            mapdMain.redrawCanvas();
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }

    /* WindowListener methods */
    public void windowClosing(WindowEvent we) {
        exitApp(0);
    }

    public void windowOpened(WindowEvent we) {
    }

    public void windowClosed(WindowEvent we) {
    }

    public void windowActivated(WindowEvent we) {
    }

    public void windowDeactivated(WindowEvent we) {
    }

    public void windowIconified(WindowEvent we) {
    }

    public void windowDeiconified(WindowEvent we) {
    }

    /* MouseListener methods */
    public void mousePressed(MouseEvent me) {
        if (mapdMain.isLookModeOn()) {
            if (mapdMain.getLookChar() != MapCanvas.NOCHAR) {
                ActionEvent aeChar = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, Globals.CMD_EDITCHR + mapdMain.getLookChar());
                this.actionPerformed(aeChar);
                ActionEvent aeLook = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, Globals.CMD_LOOK);
                this.actionPerformed(aeLook);
            }
        }
        if (!mapdMain.getHotCell().equals(MapCanvas.PT_OFFGRID)) {
            mapdMain.requestFocus();
        }
        updateComponents();
    }

    public void mouseReleased(MouseEvent me) {
    }

    public void mouseClicked(MouseEvent me) {
        updateComponents();
    }

    public void mouseEntered(MouseEvent me) {
    }

    public void mouseExited(MouseEvent me) {
    }

    /* MouseMotionListener methods */
    public void mouseMoved(MouseEvent me) {
        if (mapdMain.isLookModeOn()) {
            if (mapdMain.getLookChar() != MapCanvas.NOCHAR) {
                ActionEvent aeChar = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, Globals.CMD_EDITCHR + mapdMain.getLookChar());
                this.actionPerformed(aeChar);
                mapdMain.requestFocus();
            }
        }
        mapdMain.updateComponents();
    }

    public void mouseDragged(MouseEvent me) {
        if (mapdMain.isLookModeOn()) {
            if (mapdMain.getLookChar() != MapCanvas.NOCHAR) {
                ActionEvent aeChar = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, Globals.CMD_EDITCHR + mapdMain.getLookChar());
                this.actionPerformed(aeChar);
                ActionEvent aeLook = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, Globals.CMD_LOOK);
                this.actionPerformed(aeLook);
            }
        }
        if (!mapdMain.getHotCell().equals(MapCanvas.PT_OFFGRID)) {
            mapdMain.requestFocus();
        }
        updateComponents();
    }

    /* IconProvider methods */
    public Icon getDefaultIcon() {
        return getIconForObject(new Integer(activeChar));
    }

    public Icon getIconForObject(Object obj) {
        if (obj != null) {
            try {
                return jbtnChar[((Integer) obj)].getIcon();
            } catch (Exception e) {
                Image imgTemp = this.createImage(gcChar.getGridData().length, gcChar.getGridData()[0].length);
                if (imgTemp == null) {
                    return null;
                }
                return new ImageIcon(imgTemp);
            }
        }
        Image imgTemp = this.createImage(gcChar.getGridData().length, gcChar.getGridData()[0].length);
        if (imgTemp == null) {
            return null;
        }
        return new ImageIcon(imgTemp);
    }

    /* ScreenColorListener methods */
    public void screenColorChanged(int screenColor) {
        gcChar.setColorScreen(getScreenColorPalette()[screenColor]);
        updateFontButtons();
    }

    public void undoRedoStateChanged(boolean canUndo, boolean canRedo) {
        jbtnUndo.setEnabled(canUndo);
        jbtnRedo.setEnabled(canRedo);
        updateComponents();
    }

    // Component Builder Methods ---------------------------------------------------------------/

    protected void buildColorDock() {
        if (colorMode == COLOR_MODE_GRAPHICS_1 || colorMode == COLOR_MODE_BITMAP) {
            if (jpnlColorDock == null) {
                jpnlColorDock = getPanel(new GridLayout(2, 8, 0, 0));
            }
            else {
                jpnlColorDock.removeAll();
                jpnlColorDock.setLayout(new GridLayout(2, 8, 0, 0));
            }
            colorDockButtons = new DualClickButton[16];
            for (int cd = 0; cd < 16; cd++) {
                DualClickButton dbtnColorButton = getPaletteButton(Globals.CMD_CLRFORE + cd, Globals.CMD_CLRBACK + cd, TIGlobals.TI_PALETTE_OPAQUE[cd]);
                if (cd == gcChar.getColorBack()) {
                    dbtnColorButton.setText("B");
                }
                if (cd == gcChar.getColorDraw()) {
                    dbtnColorButton.setText("F");
                }
                jpnlColorDock.add(dbtnColorButton);
                colorDockButtons[cd] = dbtnColorButton;
            }
        }
        else {
            if (jpnlColorDock == null) {
                jpnlColorDock = getPanel(new BorderLayout());
            }
            else {
                jpnlColorDock.removeAll();
                jpnlColorDock.setLayout(new BorderLayout());
            }
            if (colorMode == COLOR_MODE_ECM_2) {
                ecmPalettes = new ECMPalette[16];
                for (int i = 0; i < 16; i++) {
                    ecmPalettes[i] = new ECMPalette(4, 4 * (i % 4));
                }
            }
            else {
                ecmPalettes = new ECMPalette[8];
                for (int i = 0; i < 8; i++) {
                    ecmPalettes[i] = new ECMPalette(8, 8 * (i % 2));
                }
            }
            ecmPaletteComboBox = new ECMPaletteComboBox(ecmPalettes, gcChar.getColorDraw(), gcChar.getColorBack(), this);
            jpnlColorDock.add(ecmPaletteComboBox, BorderLayout.CENTER);
            if (ecmCharPalettes == null) {
                ecmCharPalettes = new ECMPalette[TIGlobals.MAXCHAR + 1];
            }
            for (int i = 0; i < TIGlobals.MAXCHAR + 1; i++) {
                ecmCharPalettes[i] = ecmPalettes[0];
            }
        }
        if (jpnlColorDock.getParent() != null) {
            jpnlChTools.revalidate();
        }
        mapdMain.setScreenColorPalette(getScreenColorPalette());
    }

    protected JPanel getPanel(LayoutManager layout) {
        JPanel jpnlRtn = new JPanel(layout);
        jpnlRtn.setOpaque(true);
        jpnlRtn.setBackground(Globals.CLR_COMPONENTBACK);
        return jpnlRtn;
    }

    protected Image getImage(String name) {
        return Toolkit.getDefaultToolkit().getImage(name);
    }

    protected ImageIcon getIcon(String name) {
        URL imageURL = getClass().getResource("images/icon_" + name + "_mono.png");
        return new ImageIcon(Toolkit.getDefaultToolkit().getImage(imageURL));
    }

    protected JButton getToolButton(String buttonkey, String tooltip, Color bgcolor) {
        JButton jbtnTool = new JButton(getIcon(buttonkey));
        jbtnTool.setActionCommand(buttonkey);
        jbtnTool.addActionListener(this);
        jbtnTool.setToolTipText(tooltip);
        jbtnTool.setMargin(new Insets(0, 0, 0, 0));
        jbtnTool.setBackground(bgcolor);
        jbtnTool.setPreferredSize(Globals.DM_TOOL);
        return jbtnTool;
    }

    protected JButton getToolButton(String buttonkey, String tooltip) {
        return getToolButton(buttonkey, tooltip, Globals.CLR_BUTTON_NORMAL);
    }

    protected JButton getDockButton(String buttonlabel, String actcmd, Color bgcolor) {
        JButton jbtnDock = new JButton(buttonlabel);
        jbtnDock.setActionCommand(actcmd);
        jbtnDock.addActionListener(this);
        jbtnDock.setOpaque(true);
        jbtnDock.setBackground(bgcolor);
        jbtnDock.setMargin(new Insets(0, 0, 0, 0));
        jbtnDock.setPreferredSize(Globals.DM_TOOL);
        return jbtnDock;
    }

    protected DualClickButton getPaletteButton(String forecmd, String backcmd, Color bgcolor) {
        DualClickButton dbtnPal = new DualClickButton("", forecmd, backcmd, this);
        dbtnPal.addActionListener(this);
        dbtnPal.setOpaque(true);
        dbtnPal.setBackground(bgcolor);
        dbtnPal.setMargin(new Insets(0, 0, 0, 0));
        dbtnPal.setPreferredSize(Globals.DM_TOOL);
        dbtnPal.setFocusable(false);
        return dbtnPal;
    }

    protected JLabel getLabel(String text, int align, Color clrback) {
        JLabel jlblRtn = new JLabel(text, align);
        jlblRtn.setOpaque(true);
        jlblRtn.setBackground(clrback);
        return jlblRtn;
    }

    protected JLabel getLabel(String text, int align) {
        return getLabel(text, align, Globals.CLR_COMPONENTBACK);
    }

    protected void buildCharacterDock() {
        int dockFontRows = (advancedModeOn ? ADV_FONT_ROWS : BAS_FONT_ROWS);
        if (jpnlFont != null) {
            jpnlFont.removeAll();
            jpnlFont.setLayout(new GridLayout(dockFontRows, FONT_COLS + 1));
        } else {
            jpnlFont = getPanel(new GridLayout(dockFontRows, FONT_COLS + 1));
        }
        int col = 8;
        int ccount = 1;
        int lcount = 1;
        int ucount = 1;
        for (int c = TIGlobals.MINCHAR; c <= TIGlobals.MAXCHAR; c++) {
            if (c < TIGlobals.FIRSTCHAR) {
                if (advancedModeOn) {
                    if (col >= 8) {
                        jpnlFont.add(getLabel("L" + lcount + " ", JLabel.RIGHT, CLR_CHARS_LOWER));
                        lcount++;
                        col = 0;
                    }
                    jpnlFont.add(jbtnChar[c]);
                    col++;
                }
            } else if (c > TIGlobals.LASTCHAR) {
                if (advancedModeOn) {
                    if (col >= 8) {
                        jpnlFont.add(getLabel("U" + ucount + " ", JLabel.RIGHT, CLR_CHARS_UPPER));
                        ucount++;
                        col = 0;
                    }
                    jpnlFont.add(jbtnChar[c]);
                    col++;
                }
            } else {
                if (col >= 8) {
                    if (c > TIGlobals.FINALXBCHAR) {
                        jpnlFont.add(getLabel(ccount + " ", JLabel.RIGHT, CLR_CHARS_BASE2));
                    } else {
                        jpnlFont.add(getLabel(ccount + " ", JLabel.RIGHT, CLR_CHARS_BASE1));
                    }
                    ccount++;
                    col = 0;
                }
                jpnlFont.add(jbtnChar[c]);
                col++;
            }
        }
        jpnlFont.revalidate();
    }

// Tool Methods -------------------------------------------------------------/

    protected void swapCharacters() {
        CharacterSwapDialog swapper = new CharacterSwapDialog(this, this, bSwapBoth, bSwapImgs, bAllMaps, (advancedModeOn ? TIGlobals.MINCHAR : TIGlobals.FIRSTCHAR), (advancedModeOn ? TIGlobals.MAXCHAR : TIGlobals.LASTCHAR), activeChar);
        if (swapper.isOkay()) {
            // only process if both characters are not the same
            if (swapper.getBaseChar() != swapper.getSwapChar()) {
                for (int m = (swapper.doAllMaps() ? 0 : mapdMain.getCurrentMapId()); m < (swapper.doAllMaps() ? mapdMain.getMapCount() : mapdMain.getCurrentMapId() + 1); m++) {
                    int[][] arrayToProc = mapdMain.getMapData(m);
                    for (int y = 0; y < arrayToProc.length; y++) {
                        for (int x = 0; x < arrayToProc[y].length; x++) {
                            if (arrayToProc[y][x] == swapper.getBaseChar()) {
                                arrayToProc[y][x] = swapper.getSwapChar();
                            } else if (swapper.doSwapBoth() && (arrayToProc[y][x] == swapper.getSwapChar())) {
                                arrayToProc[y][x] = swapper.getBaseChar();
                            }
                        }
                    }
                }
                if (swapper.doSwapImages()) {
                    int[][] charGrid = hmCharGrids.get(swapper.getSwapChar());
                    hmCharGrids.put(swapper.getSwapChar(), hmCharGrids.get(swapper.getBaseChar()));
                    hmCharGrids.put(swapper.getBaseChar(), charGrid);
                    if (colorMode == COLOR_MODE_BITMAP) {
                        int[][] charColors = hmCharColors.get(swapper.getSwapChar());
                        hmCharColors.put(swapper.getSwapChar(), hmCharColors.get(swapper.getBaseChar()));
                        hmCharColors.put(swapper.getBaseChar(), charColors);
                    }
                    updateFontButton(swapper.getBaseChar(), false);
                    updateFontButton(swapper.getSwapChar());
                }
            }
            bSwapBoth = swapper.doSwapBoth();
            bSwapImgs = swapper.doSwapImages();
            bAllMaps = swapper.doAllMaps();
        }
    }

    protected void analyzeCharTrans() {
        if (charTransDialog != null && charTransDialog.isVisible()) {
            charTransDialog.transferFocus();
        }
        else {
            charTransDialog = new AnalyzeCharTransDialog(this, mapdMain, hmCharImages, hmCharGrids, hmCharColors, clrSets, colorMode);
        }
    }

// File Handling Methods -------------------------------------------------------------------/

    protected void openDataFile(String filePath) {
        File whatFile;
        if (filePath != null) {
            whatFile = new File(filePath);
        }
        else {
            whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.OPEN_DIALOG, FILEEXTS, "Map Data Files");
        }
        if (whatFile != null && whatFile.exists()) {
            // Check color mode
            BufferedReader br = null;
            try {
                br = new BufferedReader(new FileReader(whatFile));
                String line;
                do {
                    line = br.readLine();
                    if (line != null) {
                        if (line.startsWith(Globals.KEY_COLOR_MODE)) {
                            int fileColorMode = Integer.parseInt(line.substring(Globals.KEY_COLOR_MODE.length()));
                            if (fileColorMode != colorMode) {
                                int reply = confirmationAction(this, "Confirm Color Mode Change", "This file was saved in " + COLOR_MODES[fileColorMode] + ". Do you want to switch to that mode before loading the file?", true);
                                if (reply == JOptionPane.YES_OPTION) {
                                    switch (fileColorMode) {
                                        case COLOR_MODE_GRAPHICS_1:
                                            jmitGraphicsColorMode.doClick();
                                            break;
                                        case COLOR_MODE_BITMAP:
                                            jmitBitmapColorMode.doClick();
                                            break;
                                        case COLOR_MODE_ECM_2:
                                            jmitECM2ColorMode.doClick();
                                            break;
                                        case COLOR_MODE_ECM_3:
                                            jmitECM3ColorMode.doClick();
                                            break;
                                    }
                                }
                                else if (reply == JOptionPane.CANCEL_OPTION) {
                                    return;
                                }
                            }
                            line = null;
                        }
                    }
                } while (line != null);
            } catch (IOException e) {
                errorAction(this, "Error determining color mode", e.getMessage());
                return;
            } catch (NumberFormatException e) {
                errorAction(this, "Error determining color mode", e.getMessage());
                return;
            } finally {
                if (br != null) {
                    try {
                        br.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
            newProject();
            mapDataFile = whatFile;
            this.setTitle(APPTITLE + " : " + mapDataFile.getName());
            MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
            magIO.readDataFile(mapDataFile);
        }
        for (int cn = 0; cn < jbtnChar.length; cn++) {
            jbtnChar[cn].setIcon((ImageIcon) null);
            jbtnChar[cn].setText((cn >= TIGlobals.CHARMAPSTART) && (cn < (TIGlobals.CHARMAPSTART + TIGlobals.CHARMAP.length)) ? "" + TIGlobals.CHARMAP[cn - TIGlobals.CHARMAPSTART] : "?");
        }
        mapdMain.goToMap(0);
        updateFontButtons();
        updateComponents();
        ActionEvent aeInit = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, Globals.CMD_EDITCHR + activeChar);
        this.actionPerformed(aeInit);
    }

    protected void saveDataFile() {
        if (mapDataFile != null && mapDataFile.isFile()) {
            MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
            magIO.writeDataFile(mapDataFile);
            updateComponents();
        }
        else {
            saveDataFileAs();
        }
    }

    protected void saveDataFileAs() {
        File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.SAVE_DIALOG, FILEEXTS, "Map Data Files");
        if (whatFile != null) {
            if (!whatFile.getAbsolutePath().toLowerCase().endsWith("." + FILEEXT)) {
                whatFile = new File(whatFile.getAbsolutePath() + "." + FILEEXT);
            }
            mapDataFile = whatFile;
            this.setTitle(APPTITLE + " : " + mapDataFile.getName());
            MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
            magIO.writeDataFile(mapDataFile);
            updateComponents();
        }
    }

    protected void appendDataFile() {
        File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.OPEN_DIALOG, FILEEXTS, "Map Data Files");
        if (whatFile != null) {
            MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
            magIO.readAppendDataFile(whatFile);
        }
        updateComponents();
    }

    protected void loadDefaultCharacters() {
        updateFontButtons();
        updateFontButton(activeChar);
        updateComponents();
    }

    protected void importCharImage(boolean isColor) {
        File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.OPEN_DIALOG, IMGEXTS, "Image Files", true);
            if (whatFile != null) {
            Image charImg = getImage(whatFile.getAbsolutePath());
            BufferedImage buffImg = new BufferedImage(8 * 8, 8 * 32, BufferedImage.TYPE_3BYTE_BGR);
            if (isColor) {
                Graphics2D g2d = ((Graphics2D) (buffImg.getGraphics()));
                g2d.setColor(TIGlobals.TI_COLOR_TRANSOPAQUE);
                g2d.fillRect(0, 0, 8 * 8, 8 * 32);
                g2d.setComposite(AlphaComposite.SrcOver);
                g2d.drawImage(charImg, 0, 0, this);
                g2d.setComposite(AlphaComposite.Src);
            } else {
                buffImg.getGraphics().drawImage(charImg, 0, 0, this);
            }
            MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
            if (isColor) {
                magIO.readCharImageColor(buffImg);
            } else {
                magIO.readCharImageMono(buffImg);
            }
            if (colorMode == Magellan.COLOR_MODE_ECM_2 || colorMode == Magellan.COLOR_MODE_ECM_3) {
                ecmPaletteComboBox.setEditable(false);
                ecmPaletteComboBox.setSelectedItem(ecmCharPalettes[activeChar]);
                ecmPaletteComboBox.setEditable(true);
            }
            updateFontButtons();
        }
        updateComponents();
    }

    protected void importVramDump() {
        File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.OPEN_DIALOG, VDPEXTS, "VRAM Dump Files", true);
        int charOffset = 0;
        int mapOffset = 0;
        int colorOffset = 0;
        if (whatFile != null) {
            MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
            if (whatFile.length() == 16392) {
                boolean bitmapMode = false;
                try {
                    FileInputStream fib = new FileInputStream(whatFile);
                    fib.skip(16384);
                    int[] vdpRegs = new int[8];
                    for (int i = 0; i < 8; i++) {
                        vdpRegs[i] = fib.read();
                    }
                    bitmapMode = (vdpRegs[0] & 0x02) != 0;
                    mapOffset = (vdpRegs[2] & 0x0F) * 0x400;
                    if (bitmapMode) {
                        colorOffset = (vdpRegs[3] & 0x80) != 0 ? 0x2000 : 0;
                        charOffset = (vdpRegs[4] & 0x04) != 0 ? 0x2000 : 0;
                    } else {
                        colorOffset = (vdpRegs[3] & 0xFF) * 0x40;
                        charOffset = (vdpRegs[4] & 0x07) * 0x800;
                    }
                    int screenColor = (vdpRegs[7] & 0x0F);
                    mapdMain.setColorScreen(screenColor);
                    fib.close();
                } catch (Exception e) {
                    e.printStackTrace(System.out);
                }
                magIO.readVramDumpFile(whatFile, charOffset, mapOffset, colorOffset, bitmapMode);
            } else {
                VramImportDialog importer = new VramImportDialog(this);
                if (importer.isOkay()) {
                    charOffset = importer.getCharDataOffset();
                    mapOffset = importer.getMapDataOffset();
                    colorOffset = importer.getColorDataOffset();
                    magIO.readVramDumpFile(whatFile, charOffset, mapOffset, colorOffset, importer.isBitmapMode());
                }
            }
        }
        updateFontButtons();
        updateComponents();
    }

    protected void importMapImage() {
        File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.OPEN_DIALOG, IMGEXTS, "Image Files", true);
        if (whatFile != null) {
            try {
                MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
                magIO.readMapImageFile(whatFile);
            } catch (Exception e) {
                e.printStackTrace();
                errorAction(this, "Error importing file", e.getMessage());
            }
        }
        updateFontButtons();
        updateComponents();
        ActionEvent aeInit = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, Globals.CMD_EDITCHR + activeChar);
        this.actionPerformed(aeInit);
    }

    protected void exportDataFile(boolean withExecCode) {
        MagellanExportDialog exporter = new MagellanExportDialog(MagellanExportDialog.TYPE_BASIC, this, this, bExportComments, defStartChar, defEndChar, TIGlobals.FIRSTCHAR, (withExecCode ? TIGlobals.FINALXBCHAR : TIGlobals.LASTCHAR), bCurrentMapOnly);
        if (exporter.isOkay()) {
            File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.SAVE_DIALOG, XBEXTS, "XB Data Files");
            if (whatFile != null) {
                boolean isExtensionAdded = false;
                for (int ex = 0; ex < XBEXTS.length; ex++) {
                    if (whatFile.getAbsolutePath().toLowerCase().endsWith("." + XBEXTS[ex])) {
                        isExtensionAdded = true;
                    }
                }
                if (!isExtensionAdded) {
                    whatFile = new File(whatFile.getAbsolutePath() + "." + XBEXT);
                }
                int sChar = exporter.getStartChar();
                int eChar = exporter.getEndChar();
                int aLine = exporter.getCodeLineStart();
                int cLine = exporter.getCharLineStart();
                int mLine = exporter.getMapLineStart();
                int iLine = exporter.getLineInterval();
                bExportComments = exporter.includeComments();
                bCurrentMapOnly = exporter.currentMapOnly();
                defStartChar = Math.min(sChar, eChar);
                defEndChar = Math.max(sChar, eChar);
                MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
                magIO.writeXBDataFile(whatFile, defStartChar, defEndChar, aLine, cLine, mLine, iLine, withExecCode, bExportComments, bCurrentMapOnly);
            }
        }
        exporter.dispose();
    }

    protected void exportAssemblerFile() {
        MagellanExportDialog exporter = new MagellanExportDialog(MagellanExportDialog.TYPE_ASM, this, this, bExportComments, defStartChar, defEndChar, TIGlobals.MINCHAR, TIGlobals.MAXCHAR, bCurrentMapOnly, bIncludeCharNumbers);
        if (exporter.isOkay()) {
            File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.SAVE_DIALOG, ASMEXTS, "Assembler Source Files");
            if (whatFile != null) {
                boolean isExtensionAdded = false;
                for (int ex = 0; ex < ASMEXTS.length; ex++) {
                    if (whatFile.getAbsolutePath().toLowerCase().endsWith("." + ASMEXTS[ex])) {
                        isExtensionAdded = true;
                    }
                }
                if (!isExtensionAdded) {
                    whatFile = new File(whatFile.getAbsolutePath() + "." + ASMEXT);
                }
                int sChar = exporter.getStartChar();
                int eChar = exporter.getEndChar();
                bExportComments = exporter.includeComments();
                bIncludeCharNumbers = exporter.includeCharNumbers();
                bCurrentMapOnly = exporter.currentMapOnly();
                defStartChar = Math.min(sChar, eChar);
                defEndChar = Math.max(sChar, eChar);
                MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
                magIO.writeASMDataFile(whatFile, defStartChar, defEndChar, bExportComments, bCurrentMapOnly, bIncludeCharNumbers);
            }
        }
        exporter.dispose();
    }

    protected void exportScrollFile() {
        MagellanExportDialog exporter = new MagellanExportDialog(MagellanExportDialog.TYPE_SCROLL, this, this, bExportComments, defStartChar, defEndChar, TIGlobals.MINCHAR, TIGlobals.MAXCHAR, bCurrentMapOnly, bIncludeCharNumbers);
        if (exporter.isOkay()) {
            File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.SAVE_DIALOG, ASMEXTS, "Assembler Source Files");
            if (whatFile != null) {
                boolean isExtensionAdded = false;
                for (int ex = 0; ex < ASMEXTS.length; ex++) {
                    if (whatFile.getAbsolutePath().toLowerCase().endsWith("." + ASMEXTS[ex])) {
                        isExtensionAdded = true;
                    }
                }
                if (!isExtensionAdded) {
                    whatFile = new File(whatFile.getAbsolutePath() + "." + ASMEXT);
                }
                int sChar = exporter.getStartChar();
                int eChar = exporter.getEndChar();
                bExportComments = exporter.includeComments();
                bIncludeCharNumbers = exporter.includeCharNumbers();
                bCurrentMapOnly = exporter.currentMapOnly();
                defStartChar = Math.min(sChar, eChar);
                defEndChar = Math.max(sChar, eChar);
                MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
                try {
                    magIO.writeScrollFile(whatFile, exporter.getScrollOrientation(), exporter.isWrap(), exporter.isCompress(), exporter.isCompressWords(), bExportComments, bCurrentMapOnly, bIncludeCharNumbers, exporter.getFrames());
                } catch (Exception e) {
                    e.printStackTrace();
                    errorAction(this, "Export failed", e.getMessage());
                }
            }
        }
        exporter.dispose();
    }

    protected void exportBinaryFile() {
        MagellanExportDialog exporter = new MagellanExportDialog(MagellanExportDialog.TYPE_BINARY, this, this, bExportComments, defStartChar, defEndChar, TIGlobals.MINCHAR, TIGlobals.MAXCHAR, bCurrentMapOnly);
        if (exporter.isOkay()) {
            File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.SAVE_DIALOG, BINEXTS, "Binary Data Files");
            if (whatFile != null) {
                boolean isExtensionAdded = false;
                for (int ex = 0; ex < BINEXTS.length; ex++) {
                    if (whatFile.getAbsolutePath().toLowerCase().endsWith("." + BINEXTS[ex])) {
                        isExtensionAdded = true;
                    }
                }
                if (!isExtensionAdded) {
                    whatFile = new File(whatFile.getAbsolutePath() + "." + BINEXT);
                }
                int sChar = exporter.getStartChar();
                int eChar = exporter.getEndChar();
                boolean bIncludeColorsets = exporter.includeColorsets();
                boolean bIncludeChardata = exporter.includeChardata();
                boolean bIncludeSpritedata = exporter.includeSpritedata();
                byte chunkByte = (byte) (0 | (bIncludeColorsets ? MagellanImportExport.BIN_CHUNK_COLORS : 0) | (bIncludeChardata ? MagellanImportExport.BIN_CHUNK_CHARS : 0) | (bIncludeSpritedata ? MagellanImportExport.BIN_CHUNK_SPRITES : 0));
                bCurrentMapOnly = exporter.currentMapOnly();
                defStartChar = Math.min(sChar, eChar);
                defEndChar = Math.max(sChar, eChar);
                MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
                magIO.writeBinaryFile(whatFile, chunkByte, defStartChar, defEndChar, bCurrentMapOnly);
                updateComponents();
            }
        }
        exporter.dispose();
    }

    protected void exportBinaryMapFile() {
        File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.SAVE_DIALOG, BINEXTS, "Binary Data Files");
        if (whatFile != null) {
            boolean isExtensionAdded = false;
            for (int ex = 0; ex < BINEXTS.length; ex++) {
                if (whatFile.getAbsolutePath().toLowerCase().endsWith("." + BINEXTS[ex])) {
                    isExtensionAdded = true;
                }
            }
            if (!isExtensionAdded) {
                whatFile = new File(whatFile.getAbsolutePath() + "." + BINEXT);
            }
            MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
            magIO.writeBinaryMap(whatFile);
        }
    }

    protected void exportXBDisplayMerge() throws IOException {
        MagellanExportDialog exporter = new MagellanExportDialog(MagellanExportDialog.TYPE_XBSCRMER, this, this, bExportComments, defStartChar, defEndChar, TIGlobals.MINCHAR, TIGlobals.MAXCHAR, bCurrentMapOnly);
        if (exporter.isOkay()) {
            File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.SAVE_DIALOG, ANYS, "Screen Merge Files");
            if (whatFile != null) {
                int lineNo = exporter.getCodeLineStart();
                int dispWidth = exporter.getCharLineStart();

                if (lineNo >= 0 && lineNo <= 32710 && (dispWidth == 28 || dispWidth == 32)) {
                    int currentMap = mapdMain.getCurrentMapId();
                    int[][] map = mapdMain.getMapData(currentMap);

                    byte[] init = {7, 'T', 'I', 'F', 'I', 'L', 'E', 'S', 0, 6, (byte) 0x80, 1, (byte) 0x8B, (byte) 0xA3, 6};
                    byte[] tifiles = new byte[1664];
                    System.arraycopy(init, 0, tifiles, 0, init.length);
                    int sectorPosition = 0x80;
                    byte lineSize = 0x76;
                    int mapLine = 0;
                    int outputPos = 0;
                    int colStart = 2;
                    int colEnd = 29;
                    byte stringSize = (byte) 0x70;
                    if (dispWidth == 32) {
                        lineSize = (byte) 0x86;
                        colStart = 0;
                        colEnd = 31;
                        stringSize = (byte) 0x80;
                    }

                    for (int chunk = 0; chunk < 6; chunk++) {
                        outputPos = sectorPosition;
                        tifiles[outputPos++] = lineSize;
                        tifiles[outputPos++] = (byte) (lineNo >> 8);
                        tifiles[outputPos++] = (byte) (lineNo & 0xFF);
                        tifiles[outputPos++] = (byte) 0x93; // DATA "
                        tifiles[outputPos++] = (byte) 0xC7; //
                        tifiles[outputPos++] = stringSize;
                        if (dispWidth == 28) {
                            for (int line = 0; line < 4; line++)
                                for (int column = colStart; column <= colEnd; column++)
                                    tifiles[outputPos++] = (byte) map[mapLine + line][column];
                        } else {
                            for (int line = 0; line < 4; line++)
                                for (int column = colStart; column <= colEnd; column++)
                                    tifiles[outputPos++] = (byte) (0x60 + map[mapLine + line][column]);
                        }
                        tifiles[outputPos++] = 0; // end of string
                        lineNo += 10;
                        mapLine += 4;
                        tifiles[outputPos] = (byte) 0xFF; // end of line and sector
                        sectorPosition += 0x100;
                    }
                    tifiles[outputPos++] = (byte) 0x02; // end of file
                    tifiles[outputPos++] = (byte) 0xFF;
                    tifiles[outputPos++] = (byte) 0xFF;
                    tifiles[outputPos] = (byte) 0xFF;

                    try {
                        FileOutputStream fos = new FileOutputStream(whatFile);
                        fos.write(tifiles);
                        fos.flush();
                        fos.close();
                    } catch (Exception e) {
                        e.printStackTrace(System.out);
                    }
                }
            }
        }
        exporter.dispose();
    }

    protected void exportCharImage(boolean isColor) {
        File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.SAVE_DIALOG, IMGEXTS, "Image Files");
        if (whatFile != null) {
            boolean isExtensionAdded = false;
            for (int ex = 0; ex < IMGEXTS.length; ex++) {
                if (whatFile.getAbsolutePath().toLowerCase().endsWith("." + IMGEXTS[ex])) {
                    isExtensionAdded = true;
                }
            }
            if (!isExtensionAdded) {
                whatFile = new File(whatFile.getAbsolutePath() + "." + IMGEXT);
            }
            MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
            magIO.writeCharImage(whatFile, 8, isColor);
        }
    }

    protected void exportMapImage() {
        File whatFile = getFileFromChooser((mapDataFile != null ? mapDataFile.getAbsolutePath() : "."), JFileChooser.SAVE_DIALOG, IMGEXTS, "Image Files");
        if (whatFile != null) {
            boolean isExtensionAdded = false;
            for (int ex = 0; ex < IMGEXTS.length; ex++) {
                if (whatFile.getAbsolutePath().toLowerCase().endsWith("." + IMGEXTS[ex])) {
                    isExtensionAdded = true;
                }
            }
            if (!isExtensionAdded) {
                whatFile = new File(whatFile.getAbsolutePath() + "." + IMGEXT);
            }
            updateComponents();
            MagellanImportExport magIO = new MagellanImportExport(mapdMain, ecmPalettes, clrSets, hmCharGrids, hmCharColors, ecmCharPalettes, ecmCharTransparency, colorMode);
            magIO.writeMapImage(whatFile);
        }
    }

    private File getFileFromChooser(String startDir, int dialogType, String[] exts, String desc, boolean hasImagePreview) {
        JFileChooser jfileDialog = new JFileChooser(startDir);
        jfileDialog.setDialogType(dialogType);
        if (!exts[0].equals(ANY)) {
            jfileDialog.setFileFilter(new MutableFilter(exts, desc));
        }
        if (hasImagePreview) {
            jfileDialog.setAccessory(new ImagePreview(jfileDialog));
        }
        int optionSelected = JFileChooser.CANCEL_OPTION;
        if (dialogType == JFileChooser.OPEN_DIALOG) {
            optionSelected = jfileDialog.showOpenDialog(this);
        } else if (dialogType == JFileChooser.SAVE_DIALOG) {
            optionSelected = jfileDialog.showSaveDialog(this);
        } else // default to an OPEN_DIALOG
        {
            optionSelected = jfileDialog.showOpenDialog(this);
        }
        if (optionSelected == JFileChooser.APPROVE_OPTION) {
            return jfileDialog.getSelectedFile();
        }
        return null;
    }

    private File getFileFromChooser(String startDir, int dialogType, String[] exts, String desc) {
        return getFileFromChooser(startDir, dialogType, exts, desc, false);
    }

    private void savePreferences() {
        try {
            FileOutputStream fos = new FileOutputStream(new File(System.getProperty("user.home") + "/Magellan.prefs"));
            appProperties.setProperty("magnif", "" + mapdMain.getViewScale());
            appProperties.setProperty("textCursor", (mapdMain.showTypeCell() ? "true" : "false"));
            appProperties.setProperty("exportComments", (bExportComments ? "true" : "false"));
            appProperties.setProperty("includeCharNumbers", (bIncludeCharNumbers ? "true" : "false"));
            appProperties.setProperty("currentMapOnly", (bCurrentMapOnly ? "true" : "false"));
            appProperties.setProperty("expandCharacters", (advancedModeOn ? "true" : "false"));
            appProperties.setProperty("colorMode", Integer.toString(colorMode));
            appProperties.setProperty("showGrid", (mapdMain.isShowGrid() ? "true" : "false"));
            appProperties.setProperty("showPosition", (mapdMain.showPosIndic() ? "true" : "false"));
            appProperties.setProperty("base0Position", (mapdMain.base0Position() ? "true" : "false"));
            appProperties.setProperty("defStartChar", "" + defStartChar);
            appProperties.setProperty("defEndChar", "" + defEndChar);
            if (mapDataFile != null && !mapDataFile.isDirectory()) {
                mapDataFile = mapDataFile.getParentFile();
            }
            appProperties.setProperty("filePath", mapDataFile != null ? mapDataFile.getAbsolutePath() : "");
            appProperties.store(fos, (String) null);
            fos.flush();
            fos.close();
        } catch (IOException ioe) {
            ioe.printStackTrace(System.out);
        }
    }

// Class Utility Methods -------------------------------------------------------------------/

    protected void newProject() {
        if (colorMode == COLOR_MODE_BITMAP) {
            gcChar.setColorBack(0);
            gcChar.setColorDraw(1);
            for (int i = 0; i < colorDockButtons.length; i++) {
                colorDockButtons[i].setText(i == gcChar.getColorBack() ? "B" : (i == gcChar.getColorDraw() ? "F" : ""));
            }
        }
        else if (colorMode == COLOR_MODE_ECM_2 || colorMode == COLOR_MODE_ECM_3) {
            buildColorDock();
        }
        int charNum = TIGlobals.MINCHAR;
        for (int r = 0; r < FONT_ROWS; r++) {
            clrSets[r][Globals.INDEX_CLR_BACK] = 0;
            clrSets[r][Globals.INDEX_CLR_FORE] = 1;
            for (int c = 0; c < FONT_COLS; c++) {
                int[][] emptyGrid = new int[8][8];
                for (int y = 0; y < emptyGrid.length; y++) {
                    for (int x = 0; x < emptyGrid[y].length; x++) {
                        if (charNum >= TIGlobals.CHARMAPSTART && charNum <= TIGlobals.CHARMAPEND) {
                            emptyGrid[y][x] = hmDefaultChars.get(charNum)[y][x];
                        } else {
                            emptyGrid[y][x] = 0;
                        }
                    }
                }
                hmCharGrids.put(charNum, emptyGrid);
                if (colorMode == COLOR_MODE_GRAPHICS_1) {
                    int cellNum = (r * FONT_COLS) + c;
                    jbtnChar[cellNum].setBackground(TIGlobals.TI_PALETTE_OPAQUE[clrSets[r][Globals.INDEX_CLR_BACK]]);
                    jbtnChar[cellNum].setForeground(TIGlobals.TI_COLOR_UNUSED);
                }
                else if (colorMode == COLOR_MODE_BITMAP) {
                    int[][] emptyColors = new int[8][2];
                    for (int y = 0; y < emptyColors.length; y++) {
                        emptyColors[y][0] = 0;
                        emptyColors[y][1] = 1;
                    }
                    hmCharColors.put(charNum, emptyColors);
                }
                updateFontButton(charNum, false);
                charNum++;
            }
        }
        gcChar.resetUndoRedo();
        gcChar.clearGrid();
        mapdMain.delAllMaps();
        mapdMain.setColorScreen(0);
        mapdMain.redrawCanvas();
        mapdMain.resetUndoManager();
        mapDataFile = null;
        updateComponents();
    }

    protected int confirmationAction(Frame parent, String title, String message, boolean includeCancel) {
        return JOptionPane.showConfirmDialog(parent, message, title, includeCancel ? JOptionPane.YES_NO_CANCEL_OPTION : JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
    }

    protected void informationAction(Frame parent, String title, String message) {
        JOptionPane.showMessageDialog(parent, message, title, JOptionPane.INFORMATION_MESSAGE, new ImageIcon(getClass().getResource("images/logo.png")));
    }

    protected void errorAction(Frame parent, String title, String message) {
        JOptionPane.showMessageDialog(parent, message, title, JOptionPane.ERROR_MESSAGE);
    }

    protected void limitCharGrids(int newColors, int oldColors) {
        int factor = oldColors / newColors;
        for (int ch = TIGlobals.MINCHAR; ch <= TIGlobals.MAXCHAR; ch++) {
            int[][] charGrid = hmCharGrids.get(ch);
            if (charGrid != null) {
                for (int[] row : charGrid) {
                    for (int x = 0; x < row.length; x++) {
                        if (row[x] > newColors - 1) {
                            row[x] /= factor;
                        }
                    }
                }
            }
        }
    }

// Convenience Methods ---------------------------------------------------------------------/

    protected void updateComponents() {
        updateFontButton(activeChar, false);
        int[][] gridData = gcChar.getGridData();
        jtxtChar.setText(
            (
                Globals.getByteString(gridData)
                + (colorMode == COLOR_MODE_BITMAP ? Globals.getColorHexString(gcChar.getGridColors()) : "")
                + (colorMode == COLOR_MODE_ECM_2 || colorMode == COLOR_MODE_ECM_3 ? Globals.getByteString(gridData, 2) : "")
                + (colorMode == COLOR_MODE_ECM_3 ? Globals.getByteString(gridData, 4) : "")
            ).toUpperCase()
        );
        jtxtChar.setCaretPosition(0);
        jbtnLook.setBackground((mapdMain.isLookModeOn() ? Globals.CLR_BUTTON_ACTIVE : Globals.CLR_BUTTON_NORMAL));
        jchkTransparency.setVisible(colorMode == COLOR_MODE_ECM_2 || colorMode == COLOR_MODE_ECM_3);
        jchkTransparency.setSelected(ecmCharTransparency[activeChar]);
        mapdMain.updateComponents();
    }

    protected void updateFontButton(int charnum) {
        updateFontButton(charnum, true);
    }

    protected void updateFontButton(int charnum, boolean redrawMap) {
        int imageScale = 2;
        if (lastActiveChar != charnum && lastActiveChar != MapCanvas.NOCHAR) {
            jbtnChar[lastActiveChar].setBorder(Globals.bordButtonNormal);
        }
        lastActiveChar = charnum;
        jbtnChar[charnum].setBorder(Globals.bordButtonActive);
        jlblCharInt.setText((charnum < 100 ? " " : "") + charnum);
        jlblCharHex.setText(">" + Integer.toHexString(charnum).toUpperCase());
        int charmapIndex = charnum - TIGlobals.CHARMAPSTART;
        int cset = (int) (Math.floor(charnum / 8));
        Color screenColor = getScreenColorPalette()[mapdMain.getColorScreen()];
        jbtnChar[charnum].setBackground(screenColor);
        if (hmCharGrids.get(charnum) == null) {
            if (jbtnChar[charnum].getIcon() != null) {
                jbtnChar[charnum].setIcon(null);
                jbtnChar[charnum].setText((charmapIndex >= 0) && (charmapIndex < TIGlobals.CHARMAP.length) ? "" + TIGlobals.CHARMAP[charmapIndex] : "?");
            }
            return;
        }
        if (jbtnChar[charnum].getIcon() == null) {
            jbtnChar[charnum].setText("");
        }
        Image image = this.createImage(gcChar.getGridData().length * imageScale, gcChar.getGridData()[0].length * imageScale);
        Graphics g = image.getGraphics();
        Color[] palette = colorMode == COLOR_MODE_GRAPHICS_1 || colorMode == COLOR_MODE_BITMAP ? TIGlobals.TI_PALETTE_OPAQUE : ecmCharPalettes[charnum].getColors();
        int[][] charGrid = hmCharGrids.get(charnum);
        int[][] charColors = null;
        if (colorMode == COLOR_MODE_BITMAP) {
            charColors = hmCharColors.get(charnum);
        }
        // boolean useTransparency = colorMode == Magellan.COLOR_MODE_GRAPHICS_1 || colorMode == Magellan.COLOR_MODE_BITMAP || ecmCharTransparency[charnum];
        for (int y = 0; y < charGrid.length; y++) {
            for (int x = 0; x < charGrid[y].length; x++) {
                if (colorMode == COLOR_MODE_GRAPHICS_1) {
                    if (charGrid[y][x] == 1) {
                        int fore = clrSets[cset][Globals.INDEX_CLR_FORE];
                        g.setColor(fore != 0 ? palette[fore] : screenColor);
                    }
                    else {
                        int back = clrSets[cset][Globals.INDEX_CLR_BACK];
                        g.setColor(back != 0 ? palette[back] : screenColor);
                    }
                }
                else if (colorMode == COLOR_MODE_BITMAP) {
                    int colorIndex = charColors[y][charGrid[y][x]];
                    g.setColor(colorIndex != 0 ? palette[colorIndex] : screenColor);
                }
                else {
                    int colorIndex = charGrid[y][x];
                    g.setColor(colorIndex != 0 || !ecmCharTransparency[charnum] ? palette[colorIndex] : screenColor);
                }
                g.fillRect(x * imageScale, y * imageScale, imageScale, imageScale);
            }
        }
        // if (useTransparency) {
        //     image = makeImageTransparent(image);
        // }
        jbtnChar[charnum].setIcon(new ImageIcon(image));
        hmCharImages.put(charnum, image);
        mapdMain.setCharImage(charnum, image);
        if (Globals.getByteString(hmCharGrids.get(charnum)).equals(Globals.BLANKCHAR) && (colorMode == COLOR_MODE_GRAPHICS_1 && clrSets[cset][Globals.INDEX_CLR_BACK] == 0 || colorMode == COLOR_MODE_BITMAP && Globals.isColorGridEmpty(hmCharColors.get(charnum)))) {
            jbtnChar[charnum].setIcon(null);
            jbtnChar[charnum].setText(((charnum - TIGlobals.CHARMAPSTART) >= 0 && (charnum - TIGlobals.CHARMAPSTART) < TIGlobals.CHARMAP.length) ? "" + TIGlobals.CHARMAP[charnum - TIGlobals.CHARMAPSTART] : "?");
        }
        jbtnChar[charnum].repaint();
        if (redrawMap) {
            mapdMain.redrawCanvas();
        }
    }

    protected Image makeImageTransparent(Image imgSrc) {
        ImageProducer ipTrans = new FilteredImageSource(imgSrc.getSource(), Globals.ifTrans);
        return Toolkit.getDefaultToolkit().createImage(ipTrans);
    }

    public void updateFontButtons() {
        for (int i = TIGlobals.MINCHAR; i <= TIGlobals.MAXCHAR; i++) {
            updateFontButton(i, false);
        }
        mapdMain.redrawCanvas();
    }

    public Color[] getScreenColorPalette() {
        Color[] palette = null;
        if (colorMode == COLOR_MODE_GRAPHICS_1 || colorMode == COLOR_MODE_BITMAP) {
            palette = TIGlobals.TI_PALETTE_OPAQUE;
        }
        else if (colorMode == COLOR_MODE_ECM_2) {
            palette = new Color[16];
            System.arraycopy(ecmPalettes[0].getColors(), 0, palette, 0, 4);
            System.arraycopy(ecmPalettes[1].getColors(), 0, palette, 4, 4);
            System.arraycopy(ecmPalettes[2].getColors(), 0, palette, 8, 4);
            System.arraycopy(ecmPalettes[3].getColors(), 0, palette, 12, 4);
        }
        else if (colorMode == COLOR_MODE_ECM_3) {
            palette = new Color[16];
            System.arraycopy(ecmPalettes[0].getColors(), 0, palette, 0, 8);
            System.arraycopy(ecmPalettes[1].getColors(), 0, palette, 8, 8);
        }
        return palette;
    }

// Main Method -----------------------------------------------------------------------------/

    public static void main(String[] args) {
        SwingUtilities.invokeLater(new Magellan(args.length > 0 ? args[0] : null));
    }

    public void exitApp(int status) {
        if (status == 0) {
            savePreferences();
        }
        this.dispose();
        System.exit(status);
    }
}
